/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __NI_COMMON_H__
#define __NI_COMMON_H__

#include <functional>

#ifndef DWORD
#define DWORD uint32_t
#endif

/**
 * Common flags passed to crossgen()
*/
#define NI_FLAGS_ENABLER2R              0x0001
#define NI_FLAGS_APPNI                  0x0002
#define NI_FLAGS_COMPATIBILITY          0x0004
#define NI_FLAGS_VERBOSE                0x0008
#define NI_FLAGS_INSTRUMENT             0x1000

typedef std::function<void (std::string)> afterCreate;

typedef enum {
	NI_ERROR_NONE = 0,
	NI_ERROR_INVALID_PARAMETER = -1,
	NI_ERROR_ALREADY_EXIST = -2,
	NI_ERROR_NO_SUCH_FILE = -3,
	NI_ERROR_INVALID_PACKAGE = -4,
	NI_ERROR_NOT_SUPPORTED = -5,
	NI_ERROR_UNKNOWN = -9
} ni_error_e;

/**
 * @brief : structure which contains base directory info
 */
typedef struct NiCommonOption {
	std::string runtimeDir; /**< .NETCore Runtime directory */
	std::string tizenFXDir; /**< TizenFX directory */
	std::string extraDirs; /**< ":" seperated directories which can be set by plugins */
} NiCommonOption;

/**
 * @brief initialize NICommon
 * @param[i] options base directory info
 * @return ni_error_e 0 on success, otherwise a negative error value
 */
ni_error_e initNICommon(NiCommonOption* option);

/**
 * @brief finalize NICommon
 * @details Before re-initialize NICommon, you have to finalize current NICommon first.
 */
void finalizeNICommon();

/**
 * @brief create native images for platform DLLs (.NETCore + TizenFX)
 * @param[i] flags additional flags for the image generator
 */
void createNiPlatform(DWORD flags);

/**
 * @brief create a native image for a single DLL
 * @param[i] dllPath path to input DLL
 * @param[i] flags additional flags for the image generator
 * @return ni_error_e
 */
ni_error_e createNiDll(const std::string& dllPath, DWORD flags);

/**
 * @brief create native images for TAC DLLs
 * @param[i] rootPaths paths to TAC directories
 * @param[i] count length of rootPaths
 * @param[i] flags additional flags for the image generator
 */
void createNiUnderTAC(const std::string rootPaths[], int count, DWORD flags);

/**
 * @brief create native images for all DLLs under directories
 * @param[i] rootPaths paths to directories
 * @param[i] count length of rootPaths
 * @param[i] flags additional flags for the image generator
 */
void createNiUnderDirs(const std::string rootPaths[], int count, DWORD flags);

/**
 * @brief create native images for all DLLs in a package
 * @param[i] pkgId package ID
 * @param[i] flags additional flags for the image generator
 * @return ni_error_e
 */
ni_error_e createNiUnderPkgRoot(const std::string& pkgId, DWORD flags);

/**
 * @brief create a native image for a single dll in a package
 * @param[i] pkgId package ID
 * @param[i] flags additional flags for the image generator
 * @return ni_error_e
 */
ni_error_e createNiDllUnderPkgRoot(const std::string& pkgId, const std::string& dllPath, DWORD flags);

/**
 * @brief remove platform native images (.NETCore + TizenFX)
 */
void removeNiPlatform();

/**
 * @brief remove native images under directories
 * @param[i] rootPaths paths to directories
 * @param[i] count length of rootPaths
 */
void removeNiUnderDirs(const std::string rootPaths[], int count);

/**
 * @brief remove native images of a package
 * @param[i] pkgId package ID
 * @return ni_error_e
 */
ni_error_e removeNiUnderPkgRoot(const std::string& pkgId);

/**
 * @brief regenerate native images of all installed applications
 * @param[i] flags additional flags for the image generator
 * @return ni_error_e
 */
ni_error_e regenerateAppNI(DWORD flags);

#endif /* __NI_COMMON_H__ */
