/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __UTILS_H__
#define __UTILS_H__

#include <string>
#include <vector>
#include <functional>
#include <boost/filesystem.hpp>

#include <launcher_env.h>

#ifndef PATH_SEPARATOR
#define PATH_SEPARATOR '/'
#endif

namespace bf = boost::filesystem;
namespace bs = boost::system;

enum FSFlag : int {
  FS_NONE              = 0,
  FS_MERGE_SKIP        = (1 << 0),
  FS_MERGE_OVERWRITE   = (1 << 1),
  FS_COMMIT_COPY_FILE  = (1 << 2),
  FS_PRESERVE_OWNERSHIP_AND_PERMISSIONS = (1 << 3)
};

/**
 * @brief an iterator to the begin element in the range that compares equal to option
 * @param[in] begin element
 * @param[in] end elment
 * return return true when elements match
 */
bool cmdOptionExists(char** begin, char** end, const std::string& option);

/**
 * @brief get current executable path
 * return std::string path
 */
std::string readSelfPath();

/**
 * @brief concat path with PATH_SEPARATOR
 * @param[in] destination path
 * @param[in] source path
 * return std::string result path
 */
std::string concatPath(const std::string& path1, const std::string& path2);

/**
 * @brief get absolute Path
 * @param[in] source path
 * return std::string result path
 */
std::string absolutePath(const std::string& path);

/**
 * @brief get the directory of file
 * @param[in] source path
 * return std::string result path
 */
std::string baseName(const std::string& path);

/**
 * @brief replaces all matching substrings of the regular expression with a given replacement
 * @param[in] original string
 * @param[in] pattern to match
 * @param[in] replacement string
 * return the modified string
 */
std::string replaceAll(const std::string &str, const std::string &pattern, const std::string &replace);

/**
 * @brief get root path
 * @param[in] package id
 * @param[out] root path
 */
int getRootPath(std::string pkgId, std::string& rootPath);

/**
 * @brief get exec name
 * @param[in] package id
 * @param[out] exec name
 */
int getExecName(std::string pkgId, std::string& execName);

/**
 * @brief get metadata value
 * @param[in] package id
 * @param[in] metadata key
 * @param[out] metadata value
 */
int getMetadataValue(std::string pkgId, std::string metadataKey, std::string& metadataValue);

/**
 * @brief split path with ":" delimiter and put that in the vector
 * @param[in] source path
 * @param[out] string vector
 */
void splitPath(const std::string& path, std::vector<std::string>& out);

/**
 * @brief check file is exist
 * @param[in] source path
 * @return bool
 */
bool isFileExist(const std::string& path);

/**
 * @brief get file size
 * @param[in] source path
 * @return size of file
 */
uintptr_t getFileSize(const std::string& path);

/**
 * @brief check the file is managed assembly or not.
 * @param[in] file path
 * @return return true when the file is managed assembly.
 *         otherwise return false including native image case.
 */
bool isManagedAssembly(const std::string& filePath);

/**
 * @brief check the file is native image or not.
 * @param[in] file path
 * @return return true when the file is native image.
 */
bool isNativeImage(const std::string& filePath);

/**
 * @brief find assembly files in the directories
 * @param[in] directories
 * @param[out] ":" seperated assembly path list
 */
void assembliesInDirectory(const std::vector<std::string>& directories, std::string& tpaList);

/**
 * @brief function pointer for file reader
 */
typedef std::function<void (const std::string&, const char*)> FileReader;

/**
 * @brief scan files in the given directory and run file reader function for that
 * @param[in] directory
 * @param[in] file reader function
 * @param[in] scan depth
 */
void scanFilesInDir(const std::string& directory, FileReader reader, unsigned int depth);

/**
 * @brief update assembly file info.
 * @param[in] get path
 * @param[in] set path
 * @param[in] symbolic link
 */
void updateAssemblyInfo(const std::string& getPath, const std::string& setPath, bool isSymlink = false);

/**
 * @brief create the new directory.
 * @param[in] source path
 * @return return true when the directory was created.
 */
bool createDir(const bf::path& path);

/**
 * @brief copy the directory.
 * @param[in] path to the source directory
 * @param[in] path to the target directory
 * @param[in] filesystem flag
 * @return return true when the directory was copied.
 */
bool copyDir(const bf::path& path1, const bf::path& path2, FSFlag flags = FS_NONE);

/**
 * @brief copy the file.
 * @param[in] path to the source file
 * @param[in] path to the target file
 * @return return true when the file was copied.
 */
bool copyFile(const bf::path& path1, const bf::path& path2);

/**
 * @brief moves of renames the file or directory.
 * @param[in] path to the source file
 * @param[in] path to the target file
 * @return return true when the file was moved.
 */
bool moveFile(const bf::path& path1, const bf::path& path2);

/**
 * @brief removes the file or empty directory.
 * @param[in] source path
 * @return return true when the file was deleted.
 */
bool removeFile(const bf::path& path);

/**
 * @brief removes the file or directory and all its contents.
 * @param[in] source path
 * @return return true when the file or directory was deleted.
 */
bool removeAll(const bf::path& path);

/**
 * @brief .deps.json file parser
 * @param[in] root path
 * @param[in] exec name
 * @param[in] tpa list
 * return std::vector<std::string> parser data
 */
std::vector<std::string> depsJsonParser(std::string rootPath, std::string execName, std::string tpaList);

#endif /* __UTILS_H__ */
