/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "ni_common.h"
#include "log.h"
#include "utils.h"

#include <cstring>
#include <vector>
#include <sstream>
#include <glib.h>
#include <pkgmgr_installer_info.h>

#ifdef  LOG_TAG
#undef  LOG_TAG
#endif
#define LOG_TAG "DOTNET_INSTALLER_PLUGIN"

typedef struct Metadata {
	const char *key;
	const char *value;
} Metadata;

extern "C" int PKGMGR_MDPARSER_PLUGIN_INSTALL(const char *pkgId, const char *appId, GList *list)
{
	GList *tag = NULL;
	bool mdValue = false;
	Metadata *mdInfo = NULL;
	tag = g_list_first(list);
	while (tag) {
		mdInfo = (Metadata*)tag->data;
		if (strcmp(mdInfo->key, AOT_METADATA_KEY) == 0 && strcmp(mdInfo->value, METADATA_VALUE) == 0) {
			_DBG("Prefer dotnet application AOT set TRUE");
			mdValue = true;
		}
		tag = g_list_next(tag);
	}

	if (mdValue) {
		NiCommonOption option = {std::string(), std::string(), std::string()};
		if (initNICommon(&option) < 0) {
			_ERR("Fail to initialize NI Common");
			return -1;
		}

		if (createNiUnderPkgRoot(pkgId, 0) != NI_ERROR_NONE) {
			_ERR("Failed to get root path from [%s]", pkgId);
			return -1;
		} else {
			_INFO("Complete make application to native image");
		}

		std::string pkgRoot;
		if (getRootPath(pkgId, pkgRoot) < 0) {
			_ERR("Failed to get root path from [%s]", pkgId);
			return 0;
		}

		std::string binDir = concatPath(pkgRoot, "bin");
		std::string tacDir = concatPath(binDir, TAC_SYMLINK_SUB_DIR);
		if (bf::exists(tacDir)) {
			uid_t uid = 0;
			if (pkgmgr_installer_info_get_target_uid(&uid) < 0) {
				_ERR("Failed to get UID");
				return 0;
			}
			try {
				for (auto& symlinkAssembly : bf::recursive_directory_iterator(tacDir)) {
					std::string symPath = symlinkAssembly.path().string();
					if (!isNativeImage(symPath)) {
						std::string originPath = bf::read_symlink(symPath).string();
						std::string originNiPath = originPath.substr(0, originPath.rfind(".dll")) + ".ni.dll";
						if (!bf::exists(originNiPath)) {
							if(createNiDllUnderPkgRoot(pkgId, originPath, 0) != NI_ERROR_NONE) {
								_ERR("Failed to create NI file [%s]", originPath.c_str());
								return -1;
							}
						}
						std::string setNiPath = symPath.substr(0, symPath.rfind(".dll")) + ".ni.dll";
						if (!bf::exists(setNiPath)) {
							bf::create_symlink(originNiPath, setNiPath);
							_INFO("%s symbolic link file generated successfully.", setNiPath.c_str());
							if (lchown(setNiPath.c_str(), uid, 0)) {
								_ERR("Failed to change owner of: %s", setNiPath.c_str());
								return -1;
							}
						}
					}
				}
			} catch (const bf::filesystem_error& error) {
				_ERR("File system error while interating files: %s", error.what());
				return -1;
			}
		}
	}
	return 0;
}

extern "C" int PKGMGR_MDPARSER_PLUGIN_UPGRADE(const char *pkgId, const char *appId, GList *list)
{
	return PKGMGR_MDPARSER_PLUGIN_INSTALL(pkgId, appId, list);
}

extern "C" int PKGMGR_MDPARSER_PLUGIN_UNINSTALL(const char *pkgId, const char *appId, GList *list)
{
	return 0;
}

extern "C" int PKGMGR_MDPARSER_PLUGIN_REMOVED(const char *pkgId, const char *appId, GList *list)
{
	return PKGMGR_MDPARSER_PLUGIN_UPGRADE(pkgId, appId, list);
}

extern "C" int PKGMGR_MDPARSER_PLUGIN_CLEAN(const char *pkgId, const char *appId, GList *list)
{
	return 0;
}

extern "C" int PKGMGR_MDPARSER_PLUGIN_UNDO(const char *pkgId, const char *appId, GList *list)
{
	return 0;
}
