/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <dlfcn.h>
#include <signal.h>

#include <string>
#include <fstream>
#include <vector>
#include <sstream>

#include <locale>
#include <codecvt>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <linux/limits.h>

#include <storage.h>
#include <app_common.h>

#include "injection.h"
#include "utils.h"
#include "log.h"
#include "launcher.h"
#include "dotnet_launcher.h"
#include "plugin_manager.h"
#include "path_manager.h"
#include "log_manager.h"

namespace tizen {
namespace runtime {
namespace dotnetcore {

#if defined (__aarch64__)
#define ARCHITECTURE_IDENTIFIER "arm64"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-arm64", "linux", "unix-arm64", "unix", "any", "base"};

#elif defined (__arm__)
#define ARCHITECTURE_IDENTIFIER "arm"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"tizen.5.0.0-armel", "tizen.5.0.0", "tizen.4.0.0-armel", "tizen.4.0.0", "tizen-armel", "tizen", "linux-armel", "linux", "unix-armel", "unix", "any", "base"};

#elif defined (__x86_64__)
#define ARCHITECTURE_IDENTIFIER "x64"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-x64", "linux", "unix-x64", "unix", "any", "base"};

#elif defined (__i386__)
#define ARCHITECTURE_IDENTIFIER "x86"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"tizen.5.0.0-x86", "tizen.5.0.0", "tizen.4.0.0-x86", "tizen.4.0.0", "tizen-x86", "tizen", "linux-x86", "linux", "unix-x86", "unix", "any", "base"};

#else
#error "Unknown target"
#endif

static std::string getExtraNativeLibDirs(const std::string& appRoot)
{
	std::string candidate;
	for (unsigned int i = 0; i < RID_FALLBACK_GRAPH.size(); i++) {
		if(!candidate.empty()) {
			candidate += ":";
		}
		candidate += concatPath(appRoot, "bin/runtimes/" + RID_FALLBACK_GRAPH[i] + "/native");
	}

	candidate = candidate + ":" + concatPath(appRoot, "lib/" ARCHITECTURE_IDENTIFIER);
	if (!strncmp(ARCHITECTURE_IDENTIFIER, "arm64", 5)) {
		candidate = candidate + ":" + concatPath(appRoot, "lib/aarch64");
	}

	return candidate;
}


static std::vector<std::string> __envList;

static void setEnvFromFile()
{
	std::string envList;
	std::ifstream inFile(ENV_FILE_PATH);

	__envList.clear();

	if (inFile) {
		_INFO("coreclr_env.list is found");
		inFile >> envList;

		std::istringstream ss(envList);
		std::string token;

		while (std::getline(ss, token, ':')) {
			if (!token.empty()) {
				__envList.push_back(token);
			}
		}

		for (unsigned int i = 0; i < __envList.size(); i++) {
			putenv(const_cast<char *>(__envList[i].c_str()));
		}
	} else {
		_INFO("coreclr_env.list file is not found. skip");
	}
}

#define _unused(x) ((void)(x))

struct sigaction sig_abrt_new;
struct sigaction sig_abrt_old;

static bool checkOnSigabrt = false;
static bool checkOnTerminate = false;

static void onSigabrt(int signum)
{
	// use unused variable to avoid build warning
	ssize_t ret = write(STDERR_FILENO, "onSigabrt called\n", 17);

	if (checkOnTerminate) {
		ret = write(STDERR_FILENO, "onSigabrt called while terminate. go to exit\n", 45);
		_unused(ret);
		exit(0);
	}

	if (checkOnSigabrt) {
		ret = write(STDERR_FILENO, "onSigabrt called again. go to exit\n", 35);
		_unused(ret);
		exit(0);
	}

	if (hasException()) {
		ret = write(STDERR_FILENO, "******************************************************\n", 55);
		ret = write(STDERR_FILENO, "Unhandled exception is occured. check application code\n", 55);
		ret = write(STDERR_FILENO, "******************************************************\n", 55);
	}

	checkOnSigabrt = true;
	if (sigaction(SIGABRT, &sig_abrt_old, NULL) == 0) {
		if (raise(signum) < 0) {
			ret = write(STDERR_FILENO, "Fail to raise SIGABRT\n", 22);
		}
	} else {
		ret = write(STDERR_FILENO, "Fail to set original SIGABRT handler\n", 37);
	}
	_unused(ret);
}

static void registerSigHandler()
{
	sig_abrt_new.sa_handler = onSigabrt;
	if (sigemptyset(&sig_abrt_new.sa_mask) != 0) {
		_ERR("Fail to sigemptyset");
	}

	if (sigaction(SIGABRT, &sig_abrt_new, &sig_abrt_old) < 0) {
		_ERR("Fail to add sig handler");
	}
}

static bool storage_cb(int id, storage_type_e type, storage_state_e state, const char *path, void *user_data)
{
	int* tmp = (int*)user_data;
	if (type == STORAGE_TYPE_INTERNAL)
	{
		*tmp = id;
		return false;
	}

	return true;
}

static void initEnvForSpecialFolder()
{
	int storageId;
	int error;
	char *path = NULL;

	error = storage_foreach_device_supported(storage_cb, &storageId);
	if (error != STORAGE_ERROR_NONE) {
		return;
	}

	error = storage_get_directory(storageId, STORAGE_DIRECTORY_IMAGES, &path);
	if (error == STORAGE_ERROR_NONE && path != NULL) {
		setenv("XDG_PICTURES_DIR", const_cast<char *>(path), 1);
		free(path);
		path = NULL;
	}

	error = storage_get_directory(storageId, STORAGE_DIRECTORY_MUSIC, &path);
	if (error == STORAGE_ERROR_NONE && path != NULL) {
		setenv("XDG_MUSIC_DIR", const_cast<char *>(path), 1);
		free(path);
		path = NULL;
	}

	error = storage_get_directory(storageId, STORAGE_DIRECTORY_VIDEOS, &path);
	if (error == STORAGE_ERROR_NONE && path != NULL) {
		setenv("XDG_VIDEOS_DIR", const_cast<char *>(path), 1);
		free(path);
		path = NULL;
	}
}

void CoreRuntime::preloadTypes()
{
	const static std::string initDllPath = "/usr/share/dotnet.tizen/framework/Tizen.Init.dll";
	if (!isFileExist(initDllPath)) {
		_ERR("Failed to locate Tizen.Init.dll");
		return;
	}

	typedef void (*InitDelegate)();
	InitDelegate initDelegate;

	int ret = createDelegate(__hostHandle,
		__domainId,
		"Tizen.Init",
		"Tizen.Init.TypeLoader",
		"PreloadTypes",
		(void**)&initDelegate);

	if (ret < 0) {
		_ERR("Failed to create delegate for PreloadTypes (0x%08x)", ret);
	} else {
		initDelegate();
	}
}

CoreRuntime::CoreRuntime(const char* mode) :
	initializeClr(nullptr),
	executeAssembly(nullptr),
	shutdown(nullptr),
	createDelegate(nullptr),
	setEnvironmentVariable(nullptr),
	__coreclrLib(nullptr),
	__hostHandle(nullptr),
	__domainId(-1),
	fd(0),
	__initialized(false)
{
	_INFO("Constructor called!!");

	// plugin initialize should be called before start loader mainloop.
	// In case of VD plugins, attaching secure zone is done in the plugin_initialize().
	// When attaching to a secure zone, if there is a created thread, it will failed.
	// So, plugin initialize should be called before mainloop start.
	if (initializePluginManager(mode) < 0) {
		_ERR("Failed to initialize PluginManager");
	}

	if (pluginHasLogControl()) {
		__enableLogManager = false;
	} else {
		__enableLogManager = true;
	}
}

CoreRuntime::~CoreRuntime()
{
	dispose();
}

int CoreRuntime::initialize(bool standalone)
{
	// checkInjection checks dotnet-launcher run mode
	// At the moment, this mechanism is used only when the Memory Profiler is started.
	int res = checkInjection();
	if (res != 0) {
		_ERR("Failed to initnialize Memory Profiler");
		return -1;
	}
#define __XSTR(x) #x
#define __STR(x) __XSTR(x)

#ifdef NATIVE_LIB_DIR
	__nativeLibDirectory = __STR(NATIVE_LIB_DIR);
#endif

#undef __STR
#undef __XSTR

#ifdef __arm__
	// libunwind library is used to unwind stack frame, but libunwind for ARM
	// does not support ARM vfpv3/NEON registers in DWARF format correctly.
	// Therefore let's disable stack unwinding using DWARF information
	// See https://github.com/dotnet/coreclr/issues/6698
	//
	// libunwind use following methods to unwind stack frame.
	// UNW_ARM_METHOD_ALL          0xFF
	// UNW_ARM_METHOD_DWARF        0x01
	// UNW_ARM_METHOD_FRAME        0x02
	// UNW_ARM_METHOD_EXIDX        0x04
	putenv(const_cast<char *>("UNW_ARM_UNWIND_METHOD=6"));
#endif // __arm__

	// Disable debug pipes and semaphores creation in case of non-standlone mode
	if (!standalone)
		putenv(const_cast<char *>("COMPlus_EnableDiagnostics=0"));

	// Write Debug.WriteLine to stderr
	putenv(const_cast<char *>("COMPlus_DebugWriteToStdErr=1"));

#ifdef USE_DEFAULT_BASE_ADDR
	putenv(const_cast<char *>("COMPlus_UseDefaultBaseAddr=1"));
#endif // USE_DEFAULT_BASE_ADDR

	// read string from external file and set them to environment value.
	setEnvFromFile();

	// Set environment for System.Environment.SpecialFolder
	initEnvForSpecialFolder();

	if (initializePathManager(std::string(), std::string(), std::string()) < 0) {
		_ERR("Failed to initialize PathManager");
		return -1;
	}

	if (__enableLogManager) {
		if (initializeLogManager() < 0) {
			_ERR("Failed to initnialize LogManager");
			return -1;
		}

		if (redirectFD() < 0) {
			_ERR("Failed to redirect FD");
			return -1;
		}

		if (runLoggingThread() < 0) {
			_ERR("Failed to create and run logging thread to redicrect log");
			return -1;
		}
	}

	std::string libCoreclr(concatPath(getRuntimeDir(), "libcoreclr.so"));

	__coreclrLib = dlopen(libCoreclr.c_str(), RTLD_NOW | RTLD_LOCAL);
	if (__coreclrLib == nullptr) {
		char *err = dlerror();
		_ERR("dlopen failed to open libcoreclr.so with error %s", err);
		return -1;
	}

#define CORELIB_RETURN_IF_NOSYM(type, variable, name) \
	do { \
		variable = (type)dlsym(__coreclrLib, name); \
		if (variable == nullptr) { \
			_ERR(name " is not found in the libcoreclr.so"); \
			return -1; \
		} \
	} while (0)

	CORELIB_RETURN_IF_NOSYM(coreclr_initialize_ptr, initializeClr, "coreclr_initialize");
	CORELIB_RETURN_IF_NOSYM(coreclr_execute_assembly_ptr, executeAssembly, "coreclr_execute_assembly");
	CORELIB_RETURN_IF_NOSYM(coreclr_shutdown_ptr, shutdown, "coreclr_shutdown");
	CORELIB_RETURN_IF_NOSYM(coreclr_create_delegate_ptr, createDelegate, "coreclr_create_delegate");

#undef CORELIB_RETURN_IF_NOSYM

	_INFO("libcoreclr dlopen and dlsym success");

	if (!standalone)
		pluginPreload();

	fd = open("/proc/self", O_DIRECTORY);
	std::string appRoot = std::string("/proc/self/fd/") + std::to_string(fd);
	std::string appBin = concatPath(appRoot, "bin");
	std::string appLib = concatPath(appRoot, "lib");
	std::string appTac = concatPath(appBin, TAC_SYMLINK_SUB_DIR);
	std::string probePath = appBin + ":" + appLib + ":" + appTac;
	std::string NIprobePath = concatPath(appBin, APP_NI_SUB_DIR) + ":" + concatPath(appLib, APP_NI_SUB_DIR) + ":" + appTac;
	std::string tpa = getTPA();
	std::string runtimeDir = getRuntimeDir();
	std::string nativeLibPath = getExtraNativeLibDirs(appRoot) + ":" + appBin + ":" + appLib + ":" + __nativeLibDirectory + ":" + runtimeDir;
	std::string appName = std::string("dotnet-launcher-") + std::to_string(getpid());

	if (!initializeCoreClr(appName.c_str(), probePath.c_str(), NIprobePath.c_str(), nativeLibPath.c_str(), tpa.c_str())) {
		_ERR("Failed to initialize coreclr");
		return -1;
	}

	int st = createDelegate(__hostHandle, __domainId, "Dotnet.Launcher", "Dotnet.Launcher.Environment", "SetEnvironmentVariable", (void**)&setEnvironmentVariable);
	if (st < 0 || setEnvironmentVariable == nullptr) {
		_ERR("Create delegate for Dotnet.Launcher.dll -> Dotnet.Launcher.Environment -> SetEnvironmentVariable failed (0x%08x)", st);
		return -1;
	}

	__initialized = true;

	if (!standalone)
	{
		preloadTypes();		// Preload common managed code
	}

	_INFO("CoreRuntime initialize success");

	return 0;
}

bool CoreRuntime::initializeCoreClr(const char* appId,
									 const char* assemblyProbePaths,
									 const char* NIProbePaths,
									 const char* pinvokeProbePaths,
									 const char* tpaList)
{
	const char *propertyKeys[] = {
		"TRUSTED_PLATFORM_ASSEMBLIES",
		"APP_PATHS",
		"APP_NI_PATHS",
		"NATIVE_DLL_SEARCH_DIRECTORIES",
		"AppDomainCompatSwitch"
	};

	const char *propertyValues[] = {
		tpaList,
		assemblyProbePaths,
		NIProbePaths,
		pinvokeProbePaths,
		"UseLatestBehaviorWhenTFMNotSpecified"
	};

	std::string selfPath = readSelfPath();

	int st = initializeClr(selfPath.c_str(),
							appId,
							sizeof(propertyKeys) / sizeof(propertyKeys[0]),
							propertyKeys,
							propertyValues,
							&__hostHandle,
							&__domainId);

	if (st < 0) {
		_ERR("initialize core clr fail! (0x%08x)", st);
		return false;
	}

	pluginSetCoreclrInfo(__hostHandle, __domainId, createDelegate);

	_INFO("Initialize core clr success");
	return true;
}

void CoreRuntime::dispose()
{
	// call plugin finalize function to notify finalize to plugin
	// dlclose shoud be done after coreclr shutdown to avoid breaking signal chain
	pluginFinalize();

	// ignore the signal generated by an exception that occurred during shutdown
	checkOnTerminate = true;

	if (__hostHandle != nullptr) {
		int st = shutdown(__hostHandle, __domainId);
		if (st < 0)
			_ERR("shutdown core clr fail! (0x%08x)", st);
		__hostHandle = nullptr;
	}

	if (__coreclrLib != nullptr) {
		if (dlclose(__coreclrLib) != 0) {
			_ERR("libcoreclr.so close failed");
		}

		__coreclrLib = nullptr;
	}

	finalizePluginManager();
	finalizePathManager();

	__envList.clear();

	_INFO("Dotnet runtime disposed");
}

int CoreRuntime::launch(const char* appId, const char* root, const char* path, int argc, char* argv[])
{
	if (!__initialized) {
		_ERR("Runtime is not initialized");
		return -1;
	}

	if (path == nullptr) {
		_ERR("executable path is null");
		return -1;
	}

	if (!isFileExist(path)) {
		_ERR("File not exist : %s", path);
		return -1;
	}

	if (__enableLogManager) {
		// launchpad override stdout and stderr to journalctl before launch application.
		// we have to re-override that to input pipe for logging thread.
		if (redirectFD() < 0) {
			_ERR("Failed to redirect FD");
			return -1;
		}

		registerSigHandler();
	}

	pluginSetAppInfo(appId, path);

	int fd2 = open(root, O_DIRECTORY);
	dup3(fd2, fd, O_CLOEXEC);
	if (fd2 >= 0)
		close(fd2);

	// set application data path to coreclr environment.
	// application data path can be changed by owner. So, we have to set data path just before launching.
	char* localDataPath = app_get_data_path();
	if (localDataPath != nullptr) {
		setEnvironmentVariable("XDG_DATA_HOME", localDataPath);
		free(localDataPath);
	}

	pluginBeforeExecute();

	_INFO("execute assembly : %s", path);

	unsigned int ret = 0;
	int st = executeAssembly(__hostHandle, __domainId, argc, (const char**)argv, path, &ret);
	if (st < 0)
		_ERR("Failed to Execute Assembly %s (0x%08x)", path, st);
	return ret;
}

}  // namespace dotnetcore
}  // namespace runtime
}  // namespace tizen
