/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <pkgmgr-info.h>
#include <pkgmgr_installer_info.h>
#include <aul.h>

#include "log.h"
#include "utils.h"
#include "pkgmgr_parser_plugin_interface.h"

#include <wait.h>
#include <dirent.h>
#include <sys/stat.h>

#include <algorithm>
#include <string>
#include <fstream>

#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <string.h>

#include "ni_common.h"
#include "tac_common.h"
#include "path_manager.h"
#include "plugin_manager.h"

#ifdef  LOG_TAG
#undef  LOG_TAG
#endif
#define LOG_TAG "DOTNET_INSTALLER_PLUGIN"

#ifndef CROSSGEN_PATH
#error "CROSSGEN_PATH is missed"
#endif

#define __XSTR(x) #x
#define __STR(x) __XSTR(x)
static const char* __CROSSGEN_PATH = __STR(CROSSGEN_PATH);
static const char* __TAC_DIR = __STR(TAC_DIR);

#ifdef UNIQUE_DEFAULT_BASE_ADDR_SUPPORT
static const char* __SYSTEM_BASE_FILE = __STR(SYSTEM_BASE_FILE);
#endif

#undef __STR
#undef __XSTR

static int __interval = 0;
static std::string __tpa;

static void waitInterval()
{
	// by the recommand, ignore small value for performance.
	if (__interval > 10000) {
		fprintf(stderr, "sleep %d usec\n", __interval);
		usleep(__interval);
	}
}

static std::string getNiFilePath(const std::string& dllPath)
{
	size_t index = dllPath.find_last_of(".");
	if (index == std::string::npos) {
		fprintf(stderr, "File doesnot contain extension. fail to get NI file name\n");
		return "";
	}
	std::string fName = dllPath.substr(0, index);
	std::string fExt = dllPath.substr(index, dllPath.length());

	// crossgen generate file with lower case extension only
	std::transform(fExt.begin(), fExt.end(), fExt.begin(), ::tolower);
	std::string niPath = fName + ".ni" + fExt;

	return niPath;
}

static std::string getAppNIPath(const std::string& niPath)
{
	std::string fileName;
	std::string niDirPath;
	std::string prevPath;

	size_t index = niPath.find_last_of("/");
	if (index != std::string::npos) {
		prevPath = niPath.substr(0, index);
		fileName = niPath.substr(index + 1, niPath.length());
	} else {
		prevPath = ".";
		fileName = niPath;
	}

	niDirPath = concatPath(prevPath, APP_NI_SUB_DIR);

	if (!isFileExist(niDirPath)) {
		if (mkdir(niDirPath.c_str(), 0755) == 0) {
			updateAssemblyInfo(prevPath, niDirPath);
		} else {
			fprintf(stderr, "Fail to create app ni directory (%s)\n", niDirPath.c_str());
		}
	}

	return concatPath(niDirPath, fileName);
}

static bool niExist(const std::string& path)
{
	std::string f = getNiFilePath(path);
	if (f.empty()) {
		return false;
	}

	if (isFileExist(f)) {
		return true;
	}

	// native image of System.Private.CoreLib.dll should have to overwrite
	// original file to support new coreclr
	if (path.find("System.Private.CoreLib.dll") != std::string::npos) {
		std::string coreLibBackup = path + ".Backup";
		if (isFileExist(coreLibBackup)) {
			return true;
		}
	}

	return false;
}

#ifdef UNIQUE_DEFAULT_BASE_ADDR_SUPPORT
// Get next base address to be used for system ni image from file
// __SYSTEM_BASE_FILE should be checked for existance before calling this function
static uintptr_t getNextBaseAddrFromFile()
{
	FILE *pFile = fopen(__SYSTEM_BASE_FILE, "r");
	if (pFile == NULL) {
		fprintf(stderr, "Failed to open %s\n", __SYSTEM_BASE_FILE);
		return 0;
	}

	uintptr_t addr = 0;
	uintptr_t size = 0;

	while (fscanf(pFile, "%u %u", &addr, &size) != EOF) {
	}

	fclose(pFile);

	return addr + size;
}

// Get next base address to be used for system ni image
static uintptr_t getNextBaseAddr()
{
	uintptr_t baseAddr = 0;

	if (!isFileExist(__SYSTEM_BASE_FILE)) {
		// This is the starting address for all default base addresses
		baseAddr = DEFAULT_BASE_ADDR_START;
	} else {
		baseAddr = getNextBaseAddrFromFile();

		// Round to a multple of 64K (see ZapImage::CalculateZapBaseAddress in CoreCLR)
		uintptr_t BASE_ADDRESS_ALIGNMENT = 0xffff;
		baseAddr = (baseAddr + BASE_ADDRESS_ALIGNMENT) & ~BASE_ADDRESS_ALIGNMENT;
	}

	return baseAddr;
}

// Save base address of system ni image to file
static void updateBaseAddrFile(const std::string &absNiPath, uintptr_t baseAddr)
{
	uintptr_t niSize = getFileSize(absNiPath);
	if (niSize == 0) {
		fprintf(stderr, "File %s doesn't exist\n", absNiPath.c_str());
		return;
	}

	// Write new entry to the file
	FILE *pFile = fopen(__SYSTEM_BASE_FILE, "a");
	if (pFile == NULL) {
		fprintf(stderr, "Failed to open %s\n", __SYSTEM_BASE_FILE);
		return;
	}

	fprintf(pFile, "%u %u\n", baseAddr, niSize);
	fclose(pFile);
}

// check if dll is listed in TPA
static bool isTPADll(const std::string &dllPath)
{
	std::string absDllPath = absolutePath(dllPath);

	if (__tpa.find(absDllPath) != std::string::npos) {
		return true;
	}

	return false;
}
#endif

// baseAddr should be checked in file before getting here
static ni_error_e crossgen(const std::string& dllPath, const std::string& appPath, DWORD flags)
{
	if (!isFileExist(dllPath)) {
		fprintf(stderr, "dll file is not exist : %s\n", dllPath.c_str());
		return NI_ERROR_NO_SUCH_FILE;
	}

	if (!isManagedAssembly(dllPath)) {
		fprintf(stderr, "Input file is not a dll file : %s\n", dllPath.c_str());
		return NI_ERROR_INVALID_PARAMETER;
	}

	if (niExist(dllPath)) {
		fprintf(stderr, "Already ni file is exist for %s\n", dllPath.c_str());
		return NI_ERROR_ALREADY_EXIST;
	}

	std::string absDllPath = absolutePath(dllPath);
	std::string absNiPath = getNiFilePath(dllPath);
	if (absNiPath.empty()) {
		fprintf(stderr, "Fail to get ni file name\n");
		return NI_ERROR_UNKNOWN;
	}

	bool isAppNI = flags & NI_FLAGS_APPNI;
	if (isAppNI && strstr(absNiPath.c_str(), __TAC_DIR) == NULL) {
		absNiPath = getAppNIPath(absNiPath);
	}

#ifdef UNIQUE_DEFAULT_BASE_ADDR_SUPPORT
	uintptr_t baseAddr = 0;

	if (isTPADll(dllPath)) {
		baseAddr = getNextBaseAddr();
	}
#endif

	pid_t pid = fork();
	if (pid == -1)
		return NI_ERROR_UNKNOWN;

	if (pid > 0) {
		int status;
		waitpid(pid, &status, 0);
		if (WIFEXITED(status)) {
			// Do not use niExist() function to check whether ni file created or not.
			// niEixst() return false for System.Private.Corelib.dll
			if (isFileExist(absNiPath)) {
				updateAssemblyInfo(absDllPath, absNiPath);
#ifdef UNIQUE_DEFAULT_BASE_ADDR_SUPPORT
				if (baseAddr != 0) {
					updateBaseAddrFile(absNiPath, baseAddr);
				}
#endif
				return NI_ERROR_NONE;
			} else {
				fprintf(stderr, "Fail to create native image for %s\n", dllPath.c_str());
				return NI_ERROR_NO_SUCH_FILE;
			}
		}
	} else {
		std::string jitPath = getRuntimeDir() + "/libclrjit.so";
		std::vector<const char*> argv = {
			__CROSSGEN_PATH,
			"/nologo",
			"/JITPath", jitPath.c_str()
		};

		bool compat = flags & NI_FLAGS_COMPATIBILITY;
		argv.push_back(compat ? "/Trusted_Platform_Assemblies" : "/r");
		argv.push_back(__tpa.c_str());

		bool enableR2R = flags & NI_FLAGS_ENABLER2R;
		if (!enableR2R) {
			argv.push_back("/FragileNonVersionable");
		}

		if (flags & NI_FLAGS_VERBOSE) {
			argv.push_back("/verbose");
		}

		if (flags & NI_FLAGS_INSTRUMENT) {
			argv.push_back("/Tuning");
		}

#ifdef UNIQUE_DEFAULT_BASE_ADDR_SUPPORT
		if (baseAddr != 0) {
			argv.push_back("/BaseAddress");
			argv.push_back(std::to_string(baseAddr).c_str());
		}
#endif

		argv.push_back("/App_Paths");
		std::string absAppPath;
		if (!appPath.empty()) {
			absAppPath = appPath;
		} else {
			absAppPath = baseName(absDllPath);
		}
		argv.push_back(absAppPath.c_str());

		argv.push_back("/out");
		argv.push_back(absNiPath.c_str());

		argv.push_back(absDllPath.c_str());
		argv.push_back(nullptr);

		fprintf(stderr, "+ %s (%s)\n", absDllPath.c_str(), enableR2R ? "R2R" : "FNV");

		execv(__CROSSGEN_PATH, const_cast<char* const*>(argv.data()));
		exit(0);
	}

	return NI_ERROR_NONE;
}

// callback function of "pkgmgrinfo_appinfo_metadata_filter_foreach"
static int appAotCb(pkgmgrinfo_appinfo_h handle, void *userData)
{
	char *pkgId = NULL;
	int ret = 0;
	DWORD *pFlags = (DWORD*)userData;

	ret = pkgmgrinfo_appinfo_get_pkgid(handle, &pkgId);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed to get pkgid\n");
		return -1;
	}

	if (removeNiUnderPkgRoot(pkgId) != NI_ERROR_NONE) {
		fprintf(stderr, "Failed to remove previous dlls from [%s]\n", pkgId);
		return -1;
	}

	if (resetTACPackage(pkgId) != TAC_ERROR_NONE) {
		fprintf(stderr, "Failed to remove symlink for given package [%s]\n", pkgId);
		return -1;
	}

	// Regenerate ni files with R2R mode forcibiliy. (there is no way to now which option is used)
	if (createNiUnderPkgRoot(pkgId, *pFlags) != NI_ERROR_NONE) {
		fprintf(stderr, "Failed to generate NI file [%s]\n", pkgId);
		return -1;
	} else {
		fprintf(stderr, "Complete make application to native image\n");
	}

	if (createTACPackage(pkgId, *pFlags) != TAC_ERROR_NONE) {
		fprintf(stderr, "Failed to generate symbolic link file [%s]\n", pkgId);
		return -1;
	}else {
		fprintf(stderr, "Complete make symbolic link file to tac\n");
	}

	return 0;
}

static void createCoreLibNI(DWORD flags)
{
	std::string coreLib = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll");
	std::string niCoreLib = concatPath(getRuntimeDir(), "System.Private.CoreLib.ni.dll");
	std::string coreLibBackup = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll.Backup");

	if (!isFileExist(coreLibBackup)) {

		if (!crossgen(coreLib, std::string(), flags)) {
			if (rename(coreLib.c_str(), coreLibBackup.c_str())) {
				fprintf(stderr, "Failed to rename System.Private.CoreLib.dll\n");
			}
			if (rename(niCoreLib.c_str(), coreLib.c_str())) {
				fprintf(stderr, "Failed to rename System.Private.CoreLib.ni.dll\n");
			}
		} else {
			fprintf(stderr, "Failed to create native image for %s\n", coreLib.c_str());
		}
	}
}

ni_error_e initNICommon(NiCommonOption* option)
{
#if defined(__arm__)
	// get interval value
	const char* intervalFile = "/usr/share/dotnet.tizen/lib/crossgen_interval.txt";
	std::ifstream inFile(intervalFile);
	if (inFile) {
		fprintf(stderr, "crossgen_interval.txt is found\n");
		inFile >> __interval;
	}

	if (initializePluginManager("normal")) {
		fprintf(stderr, "Fail to initialize PluginManager\n");
		return NI_ERROR_UNKNOWN;
	}
	if (initializePathManager(option->runtimeDir, option->tizenFXDir, option->extraDirs)) {
		fprintf(stderr, "Fail to initialize PathManager\n");
		return NI_ERROR_UNKNOWN;
	}

	__tpa = getTPA();

	return NI_ERROR_NONE;
#else
	fprintf(stderr, "crossgen supports arm architecture only. skip ni file generation\n");
	return NI_ERROR_NOT_SUPPORTED;
#endif
}

void finalizeNICommon()
{
	__interval = 0;

	finalizePluginManager();
	finalizePathManager();

	__tpa.clear();
}


void createNiPlatform(DWORD flags)
{
	const std::string platformDirs[] = {getRuntimeDir(), getTizenFXDir()};
	createNiUnderDirs(platformDirs, 2, flags);
}

ni_error_e createNiDll(const std::string& dllPath, DWORD flags)
{
	createCoreLibNI(flags);
	// System.Private.CoreLib.dll is generated in the createCoreLibNI function.
	// Skip if input dll is System.Private.CoreLib.dll
	if (dllPath.find("System.Private.CoreLib.dll") != std::string::npos) {
		return NI_ERROR_NONE;
	}

	ni_error_e status = crossgen(dllPath, std::string(), flags);

	return status;
}

void createNiUnderTAC(const std::string rootPaths[], int count, DWORD flags)
{
	std::string appPaths;
	try {
		for (auto& nuget : bf::recursive_directory_iterator(__TAC_DIR)) {
			std::string nugetPath = nuget.path().string();
			if (bf::is_directory(nugetPath)) {
				appPaths += nugetPath;
				appPaths += ':';
			}
		}
		if (appPaths.back() == ':') {
			appPaths.pop_back();
		}

		auto convert = [&appPaths, flags](const std::string& path, const char* name) {
			if (strstr(path.c_str(), TAC_APP_LIST_DB) != NULL ||
				strstr(path.c_str(), TAC_APP_LIST_RESTORE_DB) != NULL ||
				strstr(path.c_str(), TAC_SHA_256_INFO) != NULL)
				return;
			if (!crossgen(path, appPaths.c_str(), flags)) {
				waitInterval();
			}
		};

		for (int i = 0; i < count; i++) {
			scanFilesInDir(rootPaths[i], convert, -1);
		}
	} catch (const bf::filesystem_error& error) {
		fprintf(stderr, "Failed to recursive directory: %s", error.what());
		return;
	}
}

void createNiUnderDirs(const std::string rootPaths[], int count, DWORD flags)
{
	createCoreLibNI(flags);

	std::string appPaths;
	for (int i = 0; i < count; i++) {
		appPaths += rootPaths[i];
		appPaths += ':';
	}

	if (appPaths.back() == ':')
		appPaths.pop_back();

	std::vector<std::string> tpaAssemblies;
	splitPath(__tpa, tpaAssemblies);

	auto convert = [&appPaths, flags, tpaAssemblies](const std::string& path, const char* name) {
		bool isAppNI = flags & NI_FLAGS_APPNI;
		if (isAppNI) {
			std::string assembly = path.substr(path.rfind('/') + 1);
			bool isExist = false;
			for (auto& tpa : tpaAssemblies) {
				if (!strcmp(replaceAll(tpa, ".ni.dll", ".dll").c_str(), assembly.c_str())) {
					isExist = true;
					break;
				}
			}
			if (isExist) {
				fprintf(stderr, "%s present in the TPA list skips generation of NI file.\n", path.c_str());
				return;
			}
		}
		if (!crossgen(path, appPaths.c_str(), flags)) {
			waitInterval();
		}
	};

	for (int i = 0; i < count; i++) {
		scanFilesInDir(rootPaths[i], convert, 1);
	}

	tpaAssemblies.clear();
}

ni_error_e createNiUnderPkgRoot(const std::string& pkgId, DWORD flags)
{
	std::string pkgRoot;
	if (getRootPath(pkgId, pkgRoot) < 0) {
		return NI_ERROR_INVALID_PACKAGE;
	}

	std::string binDir = concatPath(pkgRoot, "bin");
	std::string libDir = concatPath(pkgRoot, "lib");
	std::string tacDir = concatPath(binDir, TAC_SYMLINK_SUB_DIR);
	std::string paths[] = {binDir, libDir, tacDir};

	flags |= NI_FLAGS_APPNI;
	createNiUnderDirs(paths, 3, flags);

	return NI_ERROR_NONE;
}

ni_error_e createNiDllUnderPkgRoot(const std::string& pkgId, const std::string& dllPath, DWORD flags)
{
	std::string pkgRoot;
	if (getRootPath(pkgId, pkgRoot) < 0) {
		return NI_ERROR_INVALID_PACKAGE;
	}

	std::string binDir = concatPath(pkgRoot, "bin");
	std::string libDir = concatPath(pkgRoot, "lib");
	std::string tacDir = concatPath(binDir, TAC_SYMLINK_SUB_DIR);
	std::string paths = binDir + ":" + libDir + ":" + tacDir;

	if (bf::is_symlink(dllPath)) {
		if (bf::exists(tacDir)) {
			if (!isNativeImage(dllPath)) {
				std::string originPath = bf::read_symlink(dllPath).string();
				std::string originNiPath = originPath.substr(0, originPath.rfind(".dll")) + ".ni.dll";
				if (!bf::exists(originNiPath)) {
					if(createNiDll(originPath, flags) != NI_ERROR_NONE) {
						fprintf(stderr, "Failed to create NI file [%s]\n", originPath.c_str());
						return NI_ERROR_UNKNOWN;
					}
				}
				std::string setNiPath = dllPath.substr(0, dllPath.rfind(".dll")) + ".ni.dll";
				if (!bf::exists(setNiPath)) {
					bf::create_symlink(originNiPath, setNiPath);
					fprintf(stderr, "%s symbolic link file generated successfully.\n", setNiPath.c_str());
					updateAssemblyInfo(tacDir.c_str(), setNiPath.c_str(), true);
				}
			}
		}
		return NI_ERROR_NONE;
	} else {
		std::string assembly = dllPath.substr(dllPath.rfind('/') + 1);
		std::vector<std::string> tpaAssemblies;
		splitPath(__tpa, tpaAssemblies);
		bool isExist = false;
		for (auto& tpa : tpaAssemblies) {
			if (!strcmp(replaceAll(tpa, ".ni.dll", ".dll").c_str(), assembly.c_str())) {
				isExist = true;
				break;
			}
		}
		tpaAssemblies.clear();
		if (isExist) {
			fprintf(stderr, "%s present in the TPA list skips generation of NI file.\n", dllPath.c_str());
			return NI_ERROR_NONE;
		} else {
			flags |= NI_FLAGS_APPNI;
			return crossgen(dllPath, paths, flags);
		}
	}
}

void removeNiPlatform()
{
	std::string coreLib = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll");
	std::string coreLibBackup = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll.Backup");

	if (!isFileExist(coreLibBackup)) {
		return;
	}

#ifdef UNIQUE_DEFAULT_BASE_ADDR_SUPPORT
	if (isFileExist(__SYSTEM_BASE_FILE)) {
		if (remove(__SYSTEM_BASE_FILE)) {
			fprintf(stderr, "Failed to remove %s\n", __SYSTEM_BASE_FILE);
		}
	}
#endif

	if (remove(coreLib.c_str())) {
		fprintf(stderr, "Failed to remove System.Private.CoreLib native image file\n");
	}

	if (rename(coreLibBackup.c_str(), coreLib.c_str())) {
		fprintf(stderr, "Failed to rename System.Private.CoreLib.Backup to origin\n");
	}

	const std::string platformDirs[] = {getRuntimeDir(), getTizenFXDir()};

	removeNiUnderDirs(platformDirs, 2);
}

void removeNiUnderDirs(const std::string rootPaths[], int count)
{
	auto convert = [](const std::string& path, std::string name) {
		if (isNativeImage(path)) {
			if (remove(path.c_str())) {
				fprintf(stderr, "Failed to remove %s\n", path.c_str());
			}
		}
	};

	for (int i = 0; i < count; i++)
		scanFilesInDir(rootPaths[i], convert, -1);
}

ni_error_e removeNiUnderPkgRoot(const std::string& pkgId)
{
	std::string pkgRoot;
	if (getRootPath(pkgId, pkgRoot) < 0) {
		return NI_ERROR_INVALID_PACKAGE;
	}

	std::string binDir = concatPath(pkgRoot, "bin");
	std::string libDir = concatPath(pkgRoot, "lib");
	std::string paths[] = {binDir, libDir};

	removeNiUnderDirs(paths, 2);

	std::string binNIDir = concatPath(binDir, APP_NI_SUB_DIR);
	if (isFileExist(binNIDir)) {
		if (rmdir(binNIDir.c_str()) != 0) {
			fprintf(stderr, "Failed to remove app ni dir [%s]\n", binNIDir.c_str());
		}
	}

	std::string libNIDir = concatPath(libDir, APP_NI_SUB_DIR);
	if (isFileExist(libNIDir)) {
		if (rmdir(libNIDir.c_str()) != 0) {
			fprintf(stderr, "Failed to remove app ni dir [%s]\n", libNIDir.c_str());
		}
	}

	return NI_ERROR_NONE;
}

ni_error_e regenerateAppNI(DWORD flags)
{
	int ret = 0;
	pkgmgrinfo_appinfo_metadata_filter_h handle;

	ret = pkgmgrinfo_appinfo_metadata_filter_create(&handle);
	if (ret != PMINFO_R_OK)
		return NI_ERROR_UNKNOWN;

	ret = pkgmgrinfo_appinfo_metadata_filter_add(handle, AOT_METADATA_KEY, METADATA_VALUE);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
		return NI_ERROR_UNKNOWN;
	}

	ret = pkgmgrinfo_appinfo_metadata_filter_foreach(handle, appAotCb, &flags);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed pkgmgrinfo_appinfo_metadata_filter_foreach\n");
		pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
		return NI_ERROR_UNKNOWN;
	}

	fprintf(stderr, "Success pkgmgrinfo_appinfo_metadata_filter_foreach\n");

	pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
	return NI_ERROR_NONE;
}
