/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "utils.h"
#include "ni_common.h"

#include <cstdio>
#include <cstring>

#include <algorithm>
#include <vector>

std::vector<std::string> getCmdArgs(char** begin, char** end)
{
	std::vector<std::string> list;
	for (char** itr = begin+1; itr != end; itr++) {
		if (strncmp(*itr, "--", 2) != 0)
			list.push_back(*itr);
	}
	return list;
}

static void help(const char *argv0)
{
	const char* helpDesc =
		"%s is deprecated. Please use dotnettool instead.\n"
		"Example:\n"
		"   # dotnettool --help\n\n";
	printf(helpDesc, argv0);
}

int main(int argc, char* argv[])
{
	DWORD flags = 0;
	bool pkgMode = false;
	bool dllMode = false;
	bool dirMode = false;
	bool rmPkgMode = false;
	bool pkgDllMode = false;

	NiCommonOption option = {std::string(), std::string(), std::string()};
	if (initNICommon(&option) != NI_ERROR_NONE) {
		return -1;
	}

	// Parse optional switches first.
	if (cmdOptionExists(argv, argv+argc, "--r2r")) {
		flags |= NI_FLAGS_ENABLER2R;
	}
	if (cmdOptionExists(argv, argv+argc, "--compatibility")) {
		flags |= NI_FLAGS_COMPATIBILITY;
	}
	if (cmdOptionExists(argv, argv+argc, "--instrument")) {
		flags |= NI_FLAGS_INSTRUMENT;
	}
	if (cmdOptionExists(argv, argv+argc, "--verbose")) {
		flags |= NI_FLAGS_VERBOSE;
	}

	// The following commands are mutually exclusive.
	if (cmdOptionExists(argv, argv+argc, "--help")) {
		help(argv[0]);
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--system")) {
		createNiPlatform(flags);
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--dll")) {
		dllMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--pkg")) {
		pkgMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--dir")) {
		dirMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--reset-system")) {
		removeNiPlatform();
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--reset-pkg")) {
		rmPkgMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--regen-all-app")) {
		regenerateAppNI(flags);
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--pkg-dll")) {
		pkgDllMode = true;
	} else {
		help(argv[0]);
		return 0;
	}

	std::vector<std::string> args = getCmdArgs(argv, argv+argc);

	if (args.size() < 1) {
		if (pkgMode)
			fprintf(stderr, "Package name is missing.\n");
		else if (dllMode)
			fprintf(stderr, "DLL path is missing.\n");
		help(argv[0]);
		return -1;
	}

	if (pkgMode) {
		for (const std::string pkg : args) {
			// if there is AOTed dlls under package root, that is skiped.
			int ret = createNiUnderPkgRoot(pkg, flags);
			if (ret == NI_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != NI_ERROR_NONE) {
				fprintf(stderr, "Failed to generate NI file [%s]\n", pkg.c_str());
				return -1;
			}
		}
	} else if (pkgDllMode) {
		int ret = createNiDllUnderPkgRoot(args[0], args[1], flags);
		if (ret == NI_ERROR_INVALID_PACKAGE) {
			fprintf(stderr, "Failed to get root path from [%s]\n", args[0].c_str());
			return -1;
		} else if (ret == NI_ERROR_ALREADY_EXIST) {
			// skip for already exist case
			return -1;
		} else if (ret != NI_ERROR_NONE) {
			fprintf(stderr, "Failed to generate NI file [%s]\n", args[1].c_str());
			return -1;
		}
	} else if (rmPkgMode) {
		for (const std::string pkg : args) {
			int ret = removeNiUnderPkgRoot(pkg);
			if (ret == NI_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != NI_ERROR_NONE) {
				fprintf(stderr, "Failed to remove dlls for given package [%s]\n", pkg.c_str());
				return -1;
			}
		}
	} else if (dllMode) {
		// donot return error code for generation failure.
		// we have to run crossgen for all input dlls.
		for (const std::string dll : args) {
			int ret = createNiDll(dll, flags);
			if (ret == NI_ERROR_ALREADY_EXIST) {
				// skip for already exist case
			} else if (ret != NI_ERROR_NONE) {
				fprintf(stderr, "Failed to generate NI file [%s]\n", dll.c_str());
			}
		}
	} else if (dirMode) {
		createNiUnderDirs(args.data(), args.size(), flags);
	}

	return 0;
}
